
/******************************************************************************
* MODULE     : totex.gen.cc
* DESCRIPTION: conversion of TeXmacs trees into TeX strings
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#module code_totex
#import hashmap (string, tree)

/******************************************************************************
* The totexer structure
******************************************************************************/

struct to_tex {
  string  buffer;
  int     last_space, line_count;
  bool    lf, no_lf;
  bool    latex_flag;
  hashmap<string,tree> env;

public:
  to_tex (bool latex_flag);

  void   print (string s);
  void   remove (int n=1);
  void   sep (bool text_also= FALSE, bool force= FALSE);
  void   ret ();
  string last (int n=1);

  void   print_cmd (string cmd);
  void   tex_arg (tree t, bool flag=FALSE);
  void   tex_string (string s);
  void   set (string s, tree t);
  void   assign (string s, tree t);
  void   reset (string s);
  bool   in_mode (string which);
  string get_decl (string var, string val);
  string get_cmd (string var, string val);
  void   tex_set_var (string var, string val);
  void   tex_assign (string var, tree val);
  void   tex_reset_var (string var);

  void   tex_apply (tree t);
  int    tex_special_env (string& s);
  void   tex_env (tree t);

  void   tex_script (string s, tree t);
  void   tex_table (tree t);
  void   tex (tree t);
  void   preamble (string style, string lan);
  void   postamble ();
};

/******************************************************************************
* Basic output subroutines
******************************************************************************/

void
to_tex::print (string s) {
  if (lf) {
    if (no_lf) s= "\n" * s;
    else s = "\n\n" * s;
    lf= no_lf= FALSE;
  }

  int i, n=N(s);
  for (i=0; i<n; i++) {
    // remove superfluous space
    if (N(buffer)>0) {
      if (s[i]==' ') {
	char c= buffer[N(buffer)-1];
	if (c==' ') continue;
	if (line_count==0) continue;
	if (in_mode ("math")) {
	  if ((c=='_') || (c=='^') || (c=='$') || (c=='{') || (c=='['))
	    continue;
	}
      }
      if ((s[i]==']') || (s[i]=='}')) {
	char c= buffer[N(buffer)-1];
	if ((c==' ') && in_mode ("math")) {
	  buffer[N(buffer)-1]= s[i];
	  for (last_space= N(buffer)-1; last_space>=0; last_space--)
	    if (buffer[last_space]==' ') break;
	  continue;
	}
      }
    }

    // do line breaking
    if (s[i]==' ') last_space=N(buffer);
    buffer << s[i];
    line_count++;
    if (s[i]=='\n') line_count=0;
    if ((line_count>=79) && (last_space!=-1)) {
      buffer[last_space]='\n';
      line_count=N(buffer)-last_space;
      last_space=-1;
    }
  }
}

void
to_tex::remove (int n) {
  if (N(buffer)<=n) buffer= "";
  else buffer->resize (N(buffer)-n);
}

void
to_tex::sep (bool text_also, bool force) {
  if (force) {
    buffer << " ";
    line_count++;
    return;
  }
  if (text_also || in_mode ("math"))
    if ((N(buffer)>0) && (buffer[N(buffer)-1]!=' '))
      print (" ");
}

void
to_tex::ret () {
  if (line_count>0) print ("\n");
}

string
to_tex::last (int n) {
  if (N(buffer)<n) return "";
  else return buffer (N(buffer)-n, N(buffer));
}

/******************************************************************************
* Text output
******************************************************************************/

void
to_tex::print_cmd (string cmd) { 
  if (in_mode ("math")) { sep(); print (cmd); sep(); }
  else print ("{" * cmd * "}");
}

void
to_tex::tex_arg (tree t, bool flag) {
  if (flag && is_atomic (t) && (N(t->label)==1)) print (t->label);
  else {
    print ("{");
    tex (t);
    print ("}");
  }
}

inline string
big_symbol (string s) {
  if (N(s) == 1) return s;
  return "\\" * s;
}

void
to_tex::tex_string (string s) {
  int i;
  for (i=0; i<N(s); i++)
    if (s[i]=='<') {
      i++;
      int start=i;
      while ((i<N(s)) && (s[i]!='>')) i++;
      string r= "\\" * s (start, i);
      if ((N(r)>=6) && (r (0, 6) == "\\left-"))
	r= "\\left" * big_symbol (r (6, N(r)));
      if ((N(r)>=5) && (r (0, 5) == "\\mid-"))
	r= big_symbol (r (5, N(r)));
      if ((N(r)>=7) && (r (0, 7) == "\\right-"))
	r= "\\right" * big_symbol (r (7, N(r)));
      if ((N(r)>=7) && (r (0, 5) == "\\big-"))
	r= "\\big" * big_symbol (r (5, N(r)));
      if (r == "\\big\\sum") r= "\\sum";
      if (r == "\\big\\prod") r= "\\prod";
      if (r == "\\big\\int") r= "\\int";
      if (r == "\\big\\oint") r= "\\oint";
      if (r == "\\less") r= "<";
      if (r == "\\gtr")  r= ">";
      print_cmd (r);
    }
    else {
      if (in_mode ("math")) {
	if (s[i]=='\'') { print (s[i]); continue; }
	sep ();
	if (s[i]=='*') continue;
	if (s[i]=='{') { print ("\\{"); continue; }
	if (s[i]=='}') { print ("\\}"); continue; }
	if (is_alpha (s[i])) {
	  int start=i;
	  while ((i<N(s)) && is_alpha (s[i])) i++;
	  i--;
	  if (i!=start) {
	    string ss= s (start, i+1);
	    if (command_type ["\\" * ss] == "operator")
	      print ("\\" * ss * " ");
	    else print ("\\mathrm{" * ss * "}\\;");
	    continue;
	  }
	}
	if (is_numeric (s[i])) {
	  while ((i<N(s)) && is_numeric (s[i])) {
	    print (s[i]);
	    i++;
	  }
	  i--;
	  continue;
	}
	if ((s[i]=='#') && (i<(N(s)-1)) && is_numeric (s[i+1])) {
	  print (s[i++]);
	  print (s[i]);
	  continue;
	}
      }
      print (s[i]);
    }
}
  
/******************************************************************************
* Routines concerning the environment
******************************************************************************/

void
to_tex::assign (string var, tree val) {
  if (is_func (env[var], BACKUP, 2))
    env(var)= tree (BACKUP, val, env[var][1]);
  else env(var)= val;
}

void
to_tex::set (string var, tree val) {
  env(var)= tree (BACKUP, val, env[var]);
}

void
to_tex::reset (string var) {
  if (is_func (env[var], BACKUP, 2))
    env(var)= env[var][1];
}

bool
to_tex::in_mode (string which) {
  tree t= env[MODE];
  if (is_func (t, BACKUP, 2)) t=t[0];
  return t == which;
}

string
to_tex::get_decl (string var, string val) {
  if (var==FONT_SIZE) {
    int sz= (int) ((as_double (val)*10.0) + 0.5);
    if (sz<=5) return "\\tiny";
    else if (sz<=6) return "\\scriptsize";
    else if (sz<=7) return "\\footnotesize";
    else if (sz<=9) return "\\small";
    else if (sz<=11) return "\\normalsize";
    else if (sz<=13) return "\\large";
    else if (sz<=15) return "\\Large";
    else if (sz<=18) return "\\LARGE";
    else if (sz<=22) return "\\huge";
    else return "\\Huge";
  }
  if (var==TEXT_FAMILY) {
    if (val == "rm") return "\\rmfamily";
    if (val == "ss") return "\\ssfamily";
    if (val == "tt") return "\\ttfamily";
  }
  if (var==TEXT_SERIES) {
    if (val == "medium") return "\\mdseries";
    if (val == "bold") return "\\bfseries";
  }
  if (var==TEXT_SHAPE) {
    if (val == "right") return "\\upshape";
    if (val == "slanted") return "\\slshape";
    if (val == "italic") return "\\itshape";
    if (val == "small-caps") return "\\scshape";
  }
  if (var==COLOR) return "\\" * val;
  return "";
}

string
to_tex::get_cmd (string var, string val) {
  if (var==TEXT_FAMILY) {
    if (val == "rm") return "\\textrm";
    if (val == "ss") return "\\textss";
    if (val == "tt") return "\\texttt";
  }
  if (var==TEXT_SERIES) {
    if (val == "medium") return "\\textmd";
    if (val == "bold") return "\\textbf";
  }
  if (var==TEXT_SHAPE) {
    if (val == "right") return "\\textup";
    if (val == "slanted") return "\\textsl";
    if (val == "italic") return "\\textit";
    if (val == "small-caps") return "\\textsc";
  }
  if (var==MATH_FONT) {
    if (val == "cal") return "\\mathcal";
    if (val == "Euler") return "\\mathfrak";
    if (val == "Bbb") return "\\mathbb";
  }
  if (var==MATH_FAMILY) {
    if (val == "mr") return "\\mathrm";
    if (val == "ms") return "\\mathsf";
    if (val == "mt") return "\\mathtt";
    if (val == "normal") return "\\mathnormal";
    if (val == "rm") return "\\mathrm";
    if (val == "ss") return "\\mathsf";
    if (val == "tt") return "\\mathtt";
    if (val == "bf") return "\\mathbf";
    if (val == "it") return "\\mathit";
  }
  if (var==MATH_SERIES) {
    if (val == "bold") return "\\tmmathbf";
  }
  return "";
}

static tree
arg_to_sharp (tree t, tree args, int n) {
  if (is_atomic (t)) return t;

  if (is_func (t, ARGUMENT, 1) ||
      is_func (t, APPLY, 1) ||
      is_func (t, VALUE, 1))
    {
      int i;
      for (i=0; i<n; i++)
	if (t[0] == args[i]) {
	  string s ("#0");
	  s[1]= (char) (((int) '1') + i);
	  return s;
	}
    }

  int i, nt= N(t);
  tree r (t, nt);
  for (i=0; i<nt; i++)
    r[i]= arg_to_sharp (t[i], args, n);
  return r;
}

void
to_tex::tex_assign (string var, tree val) {
  string decl;
  if (is_atomic (val)) {
    assign (var, val->label);
    decl= get_cmd (var, val->label);
  }
  if (decl == "") {
    if (is_func (val, MACRO) || is_func (val, FUNCTION)) {
      sep ();
      print ("\\newcommand{\\" * var * "}");
      if (N(val)>1) print ("[" * as_string (N(val)-1) * "]");
      print ("{");
      tex (arg_to_sharp (val[N(val)-1], val, N(val)-1));
      print ("}");
      sep ();
    }
    else {
      sep ();
      print ("\\newcommand{\\" * var * "}{");
      tex (val);
      print ("}");
      sep ();      
    }
  }
  else {
    sep (TRUE);
    print (decl);
    sep (TRUE);
  }
}

void
to_tex::tex_set_var (string var, string val) {
  set (var, val);
  if (var==MODE) {
    if (val == "text") print ("\\text{");
    if (val == "math") {
      if (last()=="$") sep (TRUE, TRUE);
      print ("$");
    }
  }
  else {
    string cmd= get_cmd (var, val);
    if (cmd == "") {
      string decl= get_decl (var, val);
      if (decl=="") print ("{");
      else print ("{" * decl * " ");
    }
    else print (cmd * "{");
  }
}

void
to_tex::tex_reset_var (string var) {
  tree t= env[var];
  if (arity (t) == 0) return;
  string val= as_string (t[0]);
  if (var==MODE) {
    if (val == "text") print ("}");
    if (val == "math") {
      if (last()=="$") sep (TRUE, TRUE);
      print ("$");
    }
  }
  else print ("}");
  reset (var);
}

/******************************************************************************
* User and system commands and environments
******************************************************************************/

bool
is_std_env (string s) {
  return
    (s == "maketitle") || (s == "abstract") ||
    (s == "theorem") || (s == "proposition") || (s == "lemma") ||
    (s == "corollary") || (s == "proof") || (s == "axiom") ||
    (s == "definition") || (s == "conjecture") ||
    (s == "remark") || (s == "note") || (s == "example") ||
    (s == "exercise") || (s == "warning") || (s == "convention") ||
    (s == "verbatim") || (s == "code") || (s == "quote") ||
    (s == "quotation") || (s == "verse") || (s == "center") ||
    (s == "indent") || (s == "body") || (s == "description") ||
    (s == "equation") || (s == "equation*") ||
    starts (s, "itemize") || starts (s, "enumerate");
}

void
to_tex::tex_apply (tree t) {
  int n= N(t);
  if (((n == 3) && (t[0] == "the-index")) ||
      ((n == 3) && (t[0] == "the-glossary")))
    return;
  if (((n == 2) && is_atomic (t[0]) && is_std_env (t[0]->label)) ||
      ((n == 5) && (t[0] == "bibliography")) ||
      ((n == 3) && (t[0] == "table-of-contents")))
    tex_env (t);
  else if ((n == 1) && (t[0]=="item")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\item");
    sep (TRUE);
  }
  else if ((n == 2) && (t[0]=="item*")) {
    no_lf= TRUE;
    sep (TRUE);
    print ("\\item["); tex (t[1]); print ("]");
    sep (TRUE);
  }
  else if ((n == 4) && (t[0] == "session")) {
    ret ();
    tex (t[3]);
    ret ();
  }
  else if ((n == 3) && (t[0] == "input")) {
    print ("\\bgroup\\ttfamily ");
    tex (t[1]);
    print ("\\egroup");
    print ("\\verb");
    tex (t[2]);
    print ("");
    ret ();
  }
  else if ((n == 2) && (t[0] == "output")) {
    print ("\\bgroup\\ttfamily ");
    tex (t[1]);
    while (buffer[N(buffer)-1] == '\n') remove (1);
    lf= FALSE;
    print ("\\egroup");
    ret ();
  }
  else if ((t[0] == "cite") || (t[0] == "nocite")) {
    sep ();
    print ("\\" * t[0]->label * "{");
    int i;
    for (i=1; i<n; i++) {
      if (i>1) {
	print (",");
	sep();
      }
      print (as_string (t[i]));
    }
    print ("}");
    sep ();
  }
  else if (n == 1) print_cmd ("\\" * as_string (t[0]));
  else {
    int i;
    sep(); print ("\\"); print (as_string (t[0]));
    for (i=1; i<n; i++) tex_arg (t[i]);
    sep();
  }
}

#define ENV_NORMAL            0
#define ENV_LIST              1
#define ENV_EQUATION          2
#define ENV_BIBLIOGRAPHY      3
#define ENV_TABLE_OF_CONTENTS 4

int
to_tex::tex_special_env (string& s) {
  if ((N(s)>=7) && (s(0,7)=="itemize")) {
    if (s == "itemize-minus") s= "itemizeminus";
    else if (s == "itemize-dot") s= "itemizedot";
    else if (s == "itemize-arrow") s= "itemizearrow";
    else s= "itemize";
    return ENV_LIST;
  }
  if ((N(s)>=9) && (s(0,9)=="enumerate")) {
    if (s == "enumerate-numeric") s= "enumeratenumeric";
    else if (s == "enumerate-roman") s= "enumerateroman";
    else if (s == "enumerate-Roman") s= "enumerateromancap";
    else if (s == "enumerate-alpha") s= "enumeratealpha";
    else if (s == "enumerate-Alpha") s= "enumeratealphacap";
    else s= "enumerate";
    return ENV_LIST;
  }
  if ((N(s)>=11) && (s(0,11)=="description")) {
    s= "description";
    return ENV_LIST;
  }

  if ((s == "equation") || (s == "equation*")) return ENV_EQUATION;

  if (s == "bibliography") return ENV_BIBLIOGRAPHY;
  if (s == "table-of-contents") return ENV_TABLE_OF_CONTENTS;

  return ENV_NORMAL;
}

void
to_tex::tex_env (tree t) {
  int i, n= N(t);
  string s= as_string (t[0]);
  int env_type= tex_special_env (s);

  switch (env_type) {
  case ENV_LIST:
    no_lf= TRUE;
    break;
  case ENV_EQUATION:
    set (MODE, "math");
    no_lf= TRUE;
    break;
  case ENV_BIBLIOGRAPHY:
    ret ();
    print ("\\bibliographystyle{" * as_string (t[2]) * "}\n");
    print ("\\bibliography{" * as_string (t[3]) * "}\n");
    return;
  case ENV_TABLE_OF_CONTENTS:
    ret ();
    print ("\\tableofcontents\n");
    return;
  }

  sep();
  print ("\\begin{" * s * "}");
  for (i=1; i<n-1; i++) tex_arg (t[i]);
  ret();
  tex (t[N(t)-1]);
  ret();
  print ("\\end{" * s * "}");
  sep();

  switch (env_type) {
  case ENV_LIST:
    no_lf= TRUE;
    break;
  case ENV_EQUATION:
    no_lf= TRUE;
    reset (MODE);
    break;
  }
}

/******************************************************************************
* mathematics
******************************************************************************/

void
to_tex::tex_script (string s, tree t) {
  bool flag;
  while (last()==" ") remove ();
  flag= (last(2)=="\\;");
  if (flag) remove (2);
  print (s);
  tex_arg (t, TRUE);
  if (flag) print ("\\;");
}

void
to_tex::tex_table (tree t) {
  bool math_mode= in_mode ("math");
  tree ft= t[1];
  tree tt= ft;
  while (is_func (tt, TABLE_FORMAT)) tt= tt[N(tt)-1];
  int nr_rows= N(tt);
  tree rt= tt[0];
  while (is_func (rt, TABLE_FORMAT)) rt= rt[N(rt)-1];
  int nr_cols= N(rt);

  string type= math_mode? string ("array"): string ("tabular");
  string bef= "", aft= "";
  bool border= FALSE;
  string def= "l";
  if (t[0] == "block") border= TRUE;
  else if (t[0] == "block*") { def= "c"; border= TRUE; }
  else if (t[0] == "tabular*") def= "c";
  else if (t[0] == "matrix") { bef= "\\left("; def= "c"; aft= "\\right)"; }
  else if (t[0] == "det") { bef= "\\left|"; def= "c"; aft= "\\right|"; }
  else if (t[0] == "choice") { bef= "\\left\\{"; def= "c"; aft= "\\right."; }
  else if (t[0] == "eqnarray*") type= "eqnarray*";
  else if (t[0] == "leqnarray*") type= "leqnarray*";

  int i, j, k;
  string fm[nr_cols];
  for (j=0; j<nr_cols; j++) fm[j]= def;
  for (k=0; k<N(ft)-1; k++)
    if (is_func (ft[k], CELL_WITH, 6) &&
	(ft[k][0] == "1") && (ft[k][1] == "-1"))
      {
	int j1= as_int (ft[k][2]);
	int j2= as_int (ft[k][3]);
	if (j1<0) j1 += nr_cols;
	else if (j1>0) j1--;
	if (j2<0) j2 += nr_cols;
	else if (j2>0) j2--;
	else j2= nr_cols-1;

	if (ft[k][4] == CELL_HALIGN)
	  for (j=j1; j<=j2; j++) {
	    tree val= ft[k][5];
	    if ((val == "l") || (val == "c") || (val == "r"))
	      fm[j]= as_string (val);
	  }
	else if ((ft[k][4] == CELL_LBORDER) ||
		 (ft[k][4] == CELL_RBORDER) ||
		 (ft[k][4] == CELL_BBORDER) ||
		 (ft[k][4] == CELL_TBORDER))
	  border= TRUE;
      }

  sep ();
  if (bef != "") { print (bef); sep (TRUE); }
  print ("\\begin{" * type * "}");
  if ((type == "array") || (type == "tabular")) {
    print ("{");
    if (border) print ("|");
    for (j=0; j<nr_cols; j++) {
      print (fm[j]);
      if (border) print ("|");
    }
    print ("}");
  }
  sep (TRUE);

  if ((type == "eqnarray*") || (type == "leqnarray*")) {
    set (MODE, "math");
    ret ();
  }
  if (border) { print ("\\hline"); sep (TRUE); }
  for (i=0; i<nr_rows; i++) {
    rt= tt[i];
    while (is_func (rt, TABLE_FORMAT)) rt= rt[N(rt)-1];
    for (j=0; j<nr_cols; j++) {
      tree ct= rt[j];
      if (is_func (ct, CELL, 1)) ct= ct[0];
      if (j>0) { sep (); print ("&"); sep (TRUE); }
      tex (ct);
    }
    print ("\\\\"); sep (TRUE);
    if ((type == "eqnarray*") || (type == "leqnarray*")) ret ();
    if (border) { print ("\\hline"); sep (TRUE); }
  }
  if ((type == "eqnarray*") || (type == "leqnarray*")) {
    reset (MODE);
    ret ();
  }

  sep (TRUE);
  print ("\\end{" * type * "}");
  if (aft != "") { sep (TRUE); print (aft); }
  sep ();
}

static bool
is_wide (tree t) {
  if (is_func (t, WIDE, 2)) return is_wide (t[0]);
  if (is_compound (t)) return TRUE;
  string s= t->label;
  int i, n= N(s);
  if (n == 0) return TRUE;
  if (n == 1) return FALSE;
  if (s[0] != '<') return TRUE;
  for (i=0; i<n; i++)
    if (s[i] == '>')
      return i != (n-1);
  return TRUE;
}

/******************************************************************************
* to_tex main routine
******************************************************************************/

static string
string_to_tex_string (string s) {
  int i, n= N(s);
  string r;
  for (i=0; i<n; i++)
    switch (s[i]) {
    case '\20': r << "``"; break;
    case '\21': r << "''"; break;
    case '\22': r << ",,"; break;
    case '\25': r << "--"; break;
    case '\26': r << "---"; break;
    case ''  : r << "SS"; break;
    case ''  : r << ""; break;
    default   : r << s[i]; break;
    }
  return r;
}

void
to_tex::tex (tree t) {
  // cout << "t] " << t << "\n";
  if (is_atomic (t)) {
    tex_string (string_to_tex_string (t->label));
    return;
  }

  int i, n= N(t);
  switch (L (t)) {
  case DOCUMENT:
    for (i=0; i<n-1; i++) {
      if (t[i] == "") no_lf= FALSE;
      else tex (t[i]);
      lf= TRUE;
    }
    tex (t[i]);
    return;
  case PARAGRAPH:
    for (i=0; i<n; i++) tex (t[i]);
    return;
  case SURROUND:
    tex (t[0]);
    tex (t[2]);
    tex (t[1]);
    return;
  case CONCAT:
    for (i=0; i<n; i++) tex (t[i]);
    return;
  case FORMAT:
    if (t[0] == NO_FIRST_INDENT) print ("\\noindent ");
    if (t[0] == LINE_BREAK) print ("\\linebreak");
    if (t[0] == NEW_LINE) print ("\n\n");
    if (t[0] == PAGE_BREAK) print ("\\pagebreak");
    if (t[0] == NEW_PAGE) print ("\\newpage");
    if (t[0] == NEXT_LINE) print ("\\\\\n");
    return;
  case HSPACE:
    if (n == 1) print ("\\hspace{" * as_string (t[0]) * "}");
    else print ("\\hspace{" * as_string (t[1]) * "}");
    return;
  case VSPACE_BEFORE:
    return;
  case VSPACE_AFTER:
    if (N(t)==1) print ("\\vspace{" * as_string (t[0]) * "}");
    else print ("\\vspace{" * as_string (t[1]) * "}");
    return;
  case SPACE:
    return;
  case HTAB:
    print ("\\hspace*{\\fill} ");
    return;
  case SPLIT:
  case MOVE:
  case RESIZE:
  case FLOAT:
  case REPEAT:
  case DECORATE_ATOMS:
  case DECORATE_LINES:
  case DECORATE_PAGES:
    return;

  case GROUP:
    print ("{");
    tex (t[0]);
    print ("}");
    return;
  case LEFT:
    tex_string ("<left-" * t[0]->label * ">");
    sep ();
    return;
  case MIDDLE:
    tex_arg (t[0]);
    sep ();
    return;
  case RIGHT:
    tex_string ("<right-" * t[0]->label * ">");
    sep ();
    return;
  case BIG:
    tex_string ("<big-" * t[0]->label * ">");
    sep ();
    return;
  case LEFT_PRIME:
    if ((last (1) == " ") || (last (1) == "\n")) remove (1);
    tex_string (t[0]->label);
    sep ();
    return;
  case RIGHT_PRIME:
    if ((last (1) == " ") || (last (1) == "\n")) remove (1);
    tex_string (t[0]->label);
    sep ();
    return;
  case BELOW:
    sep();
    print ("\\stackrel");
    tex_arg (t[0]);
    tex_arg (t[1]);
    sep();
    return;
  case ABOVE:
    sep();
    print ("\\stackrel");
    tex_arg (t[1]);
    tex_arg (t[0]);
    sep();
    return;
  case LEFT_SUB:
  case LEFT_SUP:
    /*
    sep();
    print ("\\sideset{^");
    tex_arg (t[0], TRUE);
    print ("}{}");
    */
    return;
  case RIGHT_SUB:
    tex_script ("_", t[0]);
    return;
  case RIGHT_SUP:
    tex_script ("^", t[0]);
    return;
  case FRAC:
    if (n < 2) return;
    if (latex_flag) { sep(); print ("\\frac"); }
    else print ("{");
    tex_arg (t[0]);
    if (!latex_flag) print (" \\over ");
    tex_arg (t[1]);
    if (latex_flag) sep();
    else print ("}");
    return;
  case SQRT:
    if (n == 1) {
      sep();
      print ("\\sqrt");
      tex_arg (t[0]);
      sep();
    }
    else if (n == 2) {
      sep();
      print ("\\sqrt[");
      tex (t[1]);
      print ("]");
      tex_arg (t[0]);
      sep();
    }
    return;
  case WIDE :
    if (n < 2) return;
    {
      string s= as_string (t[1]);
      sep();
      if (s == "~") {
	if (is_wide (t[0])) print ("\\widetilde");
	else print ("\\tilde");
      }
      else if (s == "<check>") print ("\\check");
      else if (s == "<vect>") print ("\\vec");
      else if (s == "<acute>") print ("\\acute");
      else if (s == "<grave>") print ("\\grave");
      else {
	if (is_wide (t[0])) print ("\\widehat");
	else print ("\\hat");
      }
      tex_arg (t[0]);
      sep();
    }
    return;
  case NEG :
    sep();
    print ("\\not");
    tex_arg (t[0]);
    sep();
    return;
  case TREE :
  case OLD_MATRIX:
  case OLD_TABLE:
  case OLD_MOSAIC:
  case OLD_MOSAIC_ITEM:
  case TABLE_FORMAT:
  case TABLE_WITH:
  case CELL_WITH:
  case TABLE_MARKER:
  case TABLE:
  case ROW:
  case CELL:
  case SUB_TABLE:
    return;

  case ASSIGN:
    if (n < 2) return;
    tex_assign (as_string (t[0]), t[1]);
    return;
  case WITH:
    for (i=0; i<n-1; i+=2)
      tex_set_var (as_string (t[i]), as_string (t[i+1]));
    tex (t[i]);
    for (i=n-1; i>=0; i-=2)
      tex_reset_var (as_string (t[i]));
    return;
  case VAR_EXPAND:
  case EXPAND:
  case APPLY:
    if ((N(t)==2) && (is_func (t[1], TABLE_FORMAT) || is_func (t[1], TABLE)))
      tex_table (t);
    else tex_apply (t);
    return;
  case MACRO:
  case FUNCTION:
  case EVAL:
    return;
  case VALUE:
    if ((N(t)==1) && is_atomic (t[0]))
      print ("{\\" * t[0]->label * "}");
    return;
  case BACKUP:
  case QUOTE:
  case DELAY:
    return;

  case OR:
  case XOR:
  case AND:
  case NOT:
  case PLUS:
  case MINUS:
  case TIMES:
  case OVER:
  case DIVIDE:
  case MODULO:
  case MERGE:
  case LENGTH:
  case RANGE:
  case NUMBER:
  case DATE:
  case TRANSLATE:
  case IS_TUPLE:
  case LOOK_UP:
  case EQUAL:
  case UNEQUAL:
  case LESS:
  case LESSEQ:
  case GREATER:
  case GREATEREQ:
  case IF:
  case CASE:
  case WHILE:
  case EXTERN:
  case AUTHORIZE:
    return;

  case INACTIVE:
  case SYMBOL:
  case LATEX:
  case HYBRID:
  case TUPLE:
  case COLLECTION:
  case ASSOCIATE:
    return;
  case LABEL:
    sep();
    print ("\\label{");
    print (as_string (t[0]));
    print ("}");
    sep();
    return;
  case REFERENCE:
    sep();
    print ("\\ref{");
    print (as_string (t[0]));
    print ("}");
    sep();
    return;
  case PAGEREF:
    sep();
    print ("\\pageref{");
    print (as_string (t[0]));
    print ("}");
    sep();
    return;
  case WRITE:
    return;
  case SPECIFIC:
    if (t[0]=="latex")
      print (tree_to_verbatim (t[1]));
    return;
  case HYPERLINK:
    sep();
    print ("\\hyperlink");
    tex_arg (t[0]);
    tex_arg (t[1]);
    sep();
    return;
  case ACTION:
    sep();
    print ("\\action");
    tex_arg (t[0]);
    tex_arg (t[1]);
    sep();
    return;
  case TAG:
  case MEANING:
    tex (t[0]);
    return;

  case GRAPHICS:
  case POINT:
  case LINE:
  case ARC:
  case BEZIER:
  case POSTSCRIPT:
    return;

  default:
    return;
  }
}

/******************************************************************************
* generation of the preamble and driver program
******************************************************************************/

void
to_tex::preamble (string style, string lan) {
  print ("\\documentclass{" * style * "}\n");

  if (lan == "czech") print ("\\usepackage[czech]{babel}\n");
  if (lan == "dutch") print ("\\usepackage[dutch]{babel}\n");
  if (lan == "french") print ("\\usepackage[french]{babel}\n");
  if (lan == "german") print ("\\usepackage[german]{babel}\n");
  if (lan == "italian") print ("\\usepackage[italian]{babel}\n");
  if (lan == "polish") print ("\\usepackage[polish]{babel}\n");
  if (lan == "portugese") print ("\\usepackage[portuges]{babel}\n");
  if (lan == "romanian") print ("\\usepackage[romanian]{babel}\n");
  if (lan == "russian")
    print ("\\usepackage[cp1251]{inputenc}\n\\usepackage[russian]{babel}\n");
  if (lan == "spanish") print ("\\usepackage[spanish]{babel}\n");
  if (lan == "swedish") print ("\\usepackage[swedish]{babel}\n");

  string path= get_env ("TEXMACS_PATH");
  while ((N(path)>0) && (path[N(path)-1]=='/'))
    path= path (0, N(path)-1);
  print ("\\IfFileExists{TeXmacs.sty}\n");
  print ("  {\\usepackage{TeXmacs}}\n");
  print ("  {\\usepackage{" * path * "/misc/latex/TeXmacs}}\n");

  if ((style == "letter") || (style == "article")) {
    print ("\\newcommand{\\chapter}[1]{");
    print ("\\tmsection{\\begin{center}\\huge #1\\end{center}}");
    print ("}\n");
  }
  if (style == "letter") {
    print ("\\newcommand{\\section}[1]{\\tmsection{\\LARGE #1}}\n");
    print ("\\newcommand{\\subsection}[1]{\\tmsection{\\Large #1}}\n");
    print ("\\newcommand{\\subsubsection}[1]{\\tmsection{\\large #1}}\n");
    print ("\\newcommand{\\paragraph}[1]{\\tmparagraph{#1}}\n");
    print ("\\newcommand{\\subparagraph}[1]{\\tmparagraph{#1}}\n");
  }

  print ("\n\\begin{document}\n");

  if (lan == "czech") print ("\\SelectCzech\n");
  if (lan == "dutch") print ("\\SelectDutch\n");
  if (lan == "french") print ("\\SelectFrench\n");
  if (lan == "german") print ("\\SelectGerman\n");
  if (lan == "italian") print ("\\SelectItalian\n");
  if (lan == "polish") print ("\\SelectPolish\n");
  if (lan == "portugese") print ("\\SelectPortugese\n");
  if (lan == "romanian") print ("\\SelectRomanian\n");
  if (lan == "russian") print ("\\SelectRussian\n");
  if (lan == "spanish") print ("\\SelectSpanish\n");
  if (lan == "swedish") print ("\\SelectSwedish\n");

  print ("\n");
}

void
to_tex::postamble () {
  print ("\n\n\\end{document}\n");
}

to_tex::to_tex (bool latex_flag2): env (UNINIT) {
  last_space= -1;
  line_count= 0;
  latex_flag= latex_flag2;
  lf        = FALSE;
  no_lf     = FALSE;
  env[MODE] = "text";
}

/******************************************************************************
* interface
******************************************************************************/

static string
tree_to_tex_latex (
  tree t, string style, string mode, string lan,
  bool latex_flag, bool document_flag)
{
  initialize_tex_conversions ();
  (void) mode;
  to_tex tt (latex_flag);
  if (document_flag) tt.preamble (style, lan);
  tt.env (MODE)= mode;
  tt.tex (t);
  if (document_flag) tt.postamble ();

  /*
  cout << "Tree to latex" << "\n";
  cout << "  t    = " << t     << "\n";
  cout << "  style= " << style << "\n";
  cout << "  mode = " << mode  << "\n";
  cout << "  lan  = " << lan   << "\n";
  cout << "---\n";
  cout << tt.buffer << "\n";
  cout << "---\n";
  */

  return tt.buffer;
}

string
tree_to_tex (tree t, string mode) {
  return tree_to_tex_latex (t, "article", mode, "english", FALSE, FALSE);
}

string
tree_to_tex_document (tree t, string style, string lan) {
  return tree_to_tex_latex (t, style, "text", lan, FALSE, TRUE);
}

string
tree_to_latex (tree t, string mode) {
  return tree_to_tex_latex (t, "article", mode, "english", TRUE, FALSE);
}

string
tree_to_latex_document (tree t, string style, string lan) {
  return tree_to_tex_latex (t, style, "text", lan, TRUE, TRUE);
}

#endmodule // code_totex
